import logging
from collections import defaultdict
from typing import Dict, List, Optional, Set

from pyspark.sql import DataFrame

from wedata.feature_store.entities.column_info import ColumnInfo
from wedata.feature_store.entities.feature import Feature
from wedata.feature_store.entities.feature_column_info import FeatureColumnInfo
from wedata.feature_store.entities.feature_lookup import FeatureLookup
from wedata.feature_store.entities.feature_spec import FeatureSpec
from wedata.feature_store.entities.feature_table import FeatureTable
from wedata.feature_store.entities.feature_table_info import FeatureTableInfo
from wedata.feature_store.entities.function_info import FunctionInfo
from wedata.feature_store.entities.on_demand_column_info import OnDemandColumnInfo
from wedata.feature_store.entities.source_data_column_info import SourceDataColumnInfo

from wedata.feature_store.utils import common_utils, validation_utils, uc_utils, schema_utils
from wedata.feature_store.utils.feature_spec_utils import assign_topological_ordering

_logger = logging.getLogger(__name__)

MAX_FEATURE_FUNCTIONS = 100


def get_features_for_tables(
        spark_client, table_names: Set[str]
) -> Dict[str, List[Feature]]:
    """
    Lookup features from the feature catalog for all table_names, return a dictionary of tablename -> list of features.
    """
    return {
        table_name: spark_client.get_features(table_name)
        for table_name in table_names
    }


def get_feature_table_metadata_for_tables(
        spark_client,
        table_names: Set[str],
) -> Dict[str, FeatureTable]:
    """
    Lookup FeatureTable metadata from the feature catalog for all table_names, return a dictionary of tablename -> FeatureTable.
    """
    feature_table_metadata = {}
    for table_name in table_names:
        feature_table_metadata[table_name] = spark_client.get_feature_table(
            table_name
        )
    return feature_table_metadata


def explode_feature_lookups(
        feature_lookups: List[FeatureLookup],
        feature_table_features_map: Dict[str, List[Feature]],
        feature_table_metadata_map: Dict[str, FeatureTable],
) -> List[FeatureColumnInfo]:
    """
    Explode FeatureLookups and collect into FeatureColumnInfos. A FeatureLookup may explode into either:
    1. A single FeatureColumnInfo, in the case where only a single feature name is specified.
    2. Multiple FeatureColumnInfos, in the cases where either multiple or all feature names are specified.

    When all features are specified in a FeatureLookup (feature_names is None),
    FeatureColumnInfos will be created for all features except primary and timestamp keys.
    The order of FeatureColumnInfos returned will be the same as the order returned by GetFeatures:
    1. All partition keys that are not primary keys, in the partition key order.
    2. All other non-key features in alphabetical order.
    """
    feature_column_infos = []
    for feature_lookup in feature_lookups:
        feature_column_infos_for_feature_lookup = _explode_feature_lookup(
            feature_lookup=feature_lookup,
            features=feature_table_features_map[feature_lookup.table_name],
            feature_table=feature_table_metadata_map[feature_lookup.table_name],
        )
        feature_column_infos += feature_column_infos_for_feature_lookup
    return feature_column_infos


def _explode_feature_lookup(
        feature_lookup: FeatureLookup,
        features: List[Feature],
        feature_table: FeatureTable,
) -> List[FeatureColumnInfo]:
    feature_names = []
    if feature_lookup._get_feature_names():
        # If the user explicitly passed in a feature name or list of feature names, use that
        feature_names.extend(feature_lookup._get_feature_names())
    else:
        # Otherwise assume the user wants all columns in the feature table
        keys = {*feature_table.primary_keys, *feature_table.timestamp_keys}
        feature_names.extend(
            [feature.name for feature in features if feature.name not in keys]
        )

    return [
        FeatureColumnInfo(
            table_name=feature_lookup.table_name,
            feature_name=feature_name,
            lookup_key=common_utils.as_list(feature_lookup.lookup_key),
            output_name=(feature_lookup._get_output_name(feature_name)),
            timestamp_lookup_key=common_utils.as_list(
                feature_lookup.timestamp_lookup_key, default=[]
            ),
        )
        for feature_name in feature_names
    ]


def load_feature_data_for_tables(
        spark_client, table_names: Set[str]
) -> Dict[str, DataFrame]:
    """
    Load feature DataFrame objects for all table_names, return a dictionary of tablename -> DataFrame.
    """
    return {
        table_name: spark_client.read_table(table_name) for table_name in table_names
    }


def _validate_and_convert_lookback_windows(
        feature_lookups: List[FeatureLookup],
) -> Dict[str, Optional[float]]:
    """
    Gets lookback_window values from all feature_lookups, validates that lookback_window values are consistent per feature table,
    converts the lookback window into total seconds, and returns a dictionary of tablename -> lookback_window values. In the
    case where lookback_window is not defined, the key value mapping will be "feature_table_name" -> None.
    """
    table_lookback_windows_map = defaultdict(set)
    for fl in feature_lookups:
        table_lookback_windows_map[fl.table_name].add(fl.lookback_window)

    for table_name, lookback_windows in table_lookback_windows_map.items():
        if len(set(lookback_windows)) > 1:
            if None in lookback_windows:
                raise ValueError(
                    f"lookback_window values must be consistently defined per feature table. '{table_name}' has "
                    f"missing lookback_window values: {lookback_windows}."
                )
            else:
                raise ValueError(
                    f"Only one value for lookback_window can be defined per feature table. '{table_name}' has "
                    f"conflicting lookback_window values: {lookback_windows}."
                )

    # convert lookback windows to seconds
    for table_name, lookback_windows in table_lookback_windows_map.items():
        # Get the only element from a single member set
        window = next(iter(lookback_windows))
        table_lookback_windows_map[table_name] = (
            window.total_seconds() if window is not None else None
        )

    return table_lookback_windows_map


def validate_feature_column_infos_data(
        spark_client,
        feature_column_infos: List[FeatureColumnInfo],
        features_by_table: Dict[str, List[Feature]],
        feature_table_data_map: Dict[str, DataFrame],
):
    """
    Validates required FeatureLookup data. Checks:
    1. Feature tables exist in Delta.
    2. Feature data types match in Delta and Feature Catalog.
    """
    table_to_features = defaultdict(list)
    for fci in feature_column_infos:
        table_to_features[fci.table_name].append(fci.feature_name)

    for table_name, features_in_spec in table_to_features.items():

        catalog_features = features_by_table[table_name]
        feature_table_data = feature_table_data_map[table_name]
        catalog_schema = {
            feature.name: feature.data_type for feature in catalog_features
        }
        delta_schema = {
            feature.name: feature.dataType
            for feature in feature_table_data.schema
        }

        for feature_name in features_in_spec:
            if feature_name not in catalog_schema:
                raise ValueError(
                    f"Unable to find feature '{feature_name}' from feature table '{table_name}' in Feature Catalog."
                )
            if feature_name not in delta_schema:
                raise ValueError(
                    f"Unable to find feature '{feature_name}' from feature table '{table_name}' in Delta."
                )


def verify_df_and_labels(
        df: DataFrame,
        label_names: List[str],
        exclude_columns: List[str],
):
    # Verify DataFrame type and column uniqueness
    validation_utils.check_dataframe_type(df)
    common_utils.validate_strings_unique(
        df.columns, "Found duplicate DataFrame column names {}."
    )

    # Validate label_names, exclude_columns are unique
    common_utils.validate_strings_unique(label_names, "Found duplicate label names {}.")
    # Verify that label_names is in DataFrame and not in exclude_columns
    for label_name in label_names:
        if label_name not in df.columns:
            raise ValueError(f"Label column '{label_name}' was not found in DataFrame")
        if label_name in exclude_columns:
            raise ValueError(f"Label column '{label_name}' cannot be excluded")


def get_uc_function_infos(
        spark_client, udf_names: Set[str]
) -> Dict[str, FunctionInfo]:
    # Note: Only GetFunction ACLs are required here. ExecuteFunction ACL will be checked at SQL execution.
    function_infos = spark_client.get_functions(list(udf_names))
    return {function_info.full_name: function_info for function_info in function_infos}


def _validate_on_demand_column_info_udfs(
        on_demand_column_infos: List[OnDemandColumnInfo],
        uc_function_infos: Dict[str, FunctionInfo],
):
    """
    Validates OnDemandColumnInfo UDFs can be applied as on-demand features. Checks:
    1. UDF is defined in Python.
    2. UDF input parameters are consistent with its input bindings.

    Note: Provided UC FunctionInfos not required by OnDemandColumnInfos are not validated.
    """
    for odci in on_demand_column_infos:
        function_info = uc_function_infos[odci.udf_name]
        if function_info.external_language != FunctionInfo.PYTHON:
            raise ValueError(
                f"FeatureFunction UDF '{odci.udf_name}' is not a Python UDF. Only Python UDFs are supported."
            )

        udf_input_params = [p.name for p in function_info.input_params]
        if odci.input_bindings.keys() != set(udf_input_params):
            raise ValueError(
                f"FeatureFunction UDF '{odci.udf_name}' input parameters {udf_input_params} "
                f"do not match input bindings {odci.input_bindings}."
            )


class _FeatureTableMetadata:
    def __init__(
            self,
            feature_table_features_map,
            feature_table_metadata_map,
            feature_table_data_map,
    ):
        self.feature_table_features_map = feature_table_features_map
        self.feature_table_metadata_map = feature_table_metadata_map
        self.feature_table_data_map = feature_table_data_map


def warn_if_non_photon_for_native_spark(use_native_spark, spark_client):
    if use_native_spark and not spark_client.is_photon_cluster():
        _logger.warning(
            "Native spark join is significantly more performant on Photon-enabled clusters. Consider "
            "switching to a Photon-enabled cluster if performance is an issue."
        )


def get_table_metadata(
        spark_client, table_names
):
    # 根据特征表名(table_names)对应的所有特征
    feature_table_features_map = get_features_for_tables(
        spark_client, table_names=table_names
    )

    feature_table_metadata_map = get_feature_table_metadata_for_tables(
        spark_client, table_names=table_names
    )

    feature_table_data_map = load_feature_data_for_tables(
        spark_client, table_names=table_names
    )
    return _FeatureTableMetadata(
        feature_table_features_map,
        feature_table_metadata_map,
        feature_table_data_map
    )


class _ColumnInfos:
    def __init__(
            self,
            source_data_column_infos,
            feature_column_infos,
            on_demand_column_infos,
    ):
        self.source_data_column_infos = source_data_column_infos
        self.feature_column_infos = feature_column_infos
        self.on_demand_column_infos = on_demand_column_infos


def get_column_infos(
        feature_lookups, feature_functions, ft_metadata, df_columns=[], label_names=[]
):
    # Collect SourceDataColumnInfos
    source_data_column_infos = [
        SourceDataColumnInfo(col) for col in df_columns if col not in label_names
    ]

    # Collect FeatureColumnInfos
    feature_column_infos = explode_feature_lookups(
        feature_lookups,
        ft_metadata.feature_table_features_map,
        ft_metadata.feature_table_metadata_map,
    )

    # Collect OnDemandColumnInfos
    on_demand_column_infos = [
        OnDemandColumnInfo(
            udf_name=feature_function.udf_name,
            input_bindings=feature_function.input_bindings,
            output_name=feature_function.output_name,
        )
        for feature_function in feature_functions
    ]
    return _ColumnInfos(
        source_data_column_infos, feature_column_infos, on_demand_column_infos
    )


def validate_column_infos(
        spark_client,
        ft_metadata,
        source_column_infos,
        feature_column_infos,
        on_demand_column_infos,
        label_names=[]
):
    source_data_names = [sdci.name for sdci in source_column_infos]

    # Verify features have unique output names
    feature_output_names = [fci.output_name for fci in feature_column_infos]
    common_utils.validate_strings_unique(
        feature_output_names, "Found duplicate feature output names {}."
    )

    # Verify labels do not collide with feature output names
    for label_name in label_names:
        if label_name in feature_output_names:
            raise ValueError(
                f"Feature cannot have same output name as label '{label_name}'."
            )

    # Verify that FeatureLookup output names do not conflict with source data names
    feature_conflicts = [
        name for name in feature_output_names if name in source_data_names
    ]
    if len(feature_conflicts) > 0:
        feature_conflicts_str = ", ".join([f"'{name}'" for name in feature_conflicts])
        raise ValueError(
            f"DataFrame contains column names that match feature output names specified"
            f" in FeatureLookups: {feature_conflicts_str}. Either remove these columns"
            f" from the DataFrame or FeatureLookups."
        )

    # Validate FeatureLookup data exists (including for columns that will be excluded).
    validate_feature_column_infos_data(
        spark_client,
        feature_column_infos,
        ft_metadata.feature_table_features_map,
        ft_metadata.feature_table_data_map,
    )

    on_demand_input_names = common_utils.get_unique_list_order(
        [
            input_name
            for odci in on_demand_column_infos
            for input_name in odci.input_bindings.values()
        ]
    )
    on_demand_output_names = [odci.output_name for odci in on_demand_column_infos]

    # Verify on-demand features have unique output names
    common_utils.validate_strings_unique(
        on_demand_output_names, "Found duplicate on-demand feature output names {}."
    )

    # Verify labels do not collide with on-demand output names
    for label_name in label_names:
        if label_name in on_demand_output_names:
            raise ValueError(
                f"On-demand feature cannot have same output name as label '{label_name}'."
            )

    # Verify on-demand feature output names do not conflict with source data or feature output names
    source_data_and_feature_output_names = set(source_data_names + feature_output_names)
    on_demand_conflicts = [
        name
        for name in on_demand_output_names
        if name in source_data_and_feature_output_names
    ]
    if len(on_demand_conflicts) > 0:
        conflicting_on_demand_feature_names = ", ".join(
            f"'{name}'" for name in on_demand_conflicts
        )
        raise ValueError(
            f"FeatureFunctions contains output names that match either DataFrame column names "
            f"or feature output names specified in FeatureLookups: {conflicting_on_demand_feature_names}. "
            f"Either remove these columns from the DataFrame, FeatureLookups, or FeatureFunctions."
        )

    # Validate on-demand and feature inputs exist in either source data or feature or function
    # outputs from previous levels
    all_output_names = source_data_and_feature_output_names.union(
        on_demand_output_names
    )
    missing_on_demand_inputs = set(on_demand_input_names).difference(all_output_names)
    if len(missing_on_demand_inputs) > 0:
        missing_on_demand_inputs_names = ", ".join(
            [f"'{name}'" for name in sorted(missing_on_demand_inputs)]
        )
        raise ValueError(
            f"Could not find input binding columns {missing_on_demand_inputs_names} required "
            "by FeatureFunctions."
        )

    feature_input_names = common_utils.get_unique_list_order(
        [input_name for fci in feature_column_infos for input_name in fci.lookup_key]
    )
    # Validate feature inputs exist in either source data or feature or function outputs
    missing_lookup_inputs = set(feature_input_names).difference(all_output_names)
    if len(missing_lookup_inputs) > 0:
        missing_input_names = ", ".join(
            [f"'{name}'" for name in sorted(missing_lookup_inputs)]
        )
        raise ValueError(
            f"Could not find lookup key columns {missing_input_names} required by "
            "FeatureLookups."
        )

    # uc_function_infos = get_uc_function_infos(
    #     spark_client,
    #     {odci.udf_name for odci in on_demand_column_infos},
    # )
    # Validate FeatureFunctions UDFs (including for columns that will be excluded).
    # _validate_on_demand_column_info_udfs(
    #     on_demand_column_infos=on_demand_column_infos,
    #     uc_function_infos=uc_function_infos,
    # )


def build_feature_spec(
        feature_lookups,
        ft_metadata,
        all_column_infos,
        exclude_columns
):
    # The order of ColumnInfos in feature_spec.yaml should be:
    # 1. SourceDataColumnInfos: non-label and non-excluded columns from the input DataFrame
    # 2. FeatureColumnInfos: features retrieved through FeatureLookups
    # 3. OnDemandColumnInfos: features created by FeatureFunctions
    column_infos = [
        ColumnInfo(info=info, include=info.output_name not in exclude_columns)
        for info in all_column_infos.source_data_column_infos
                    + all_column_infos.feature_column_infos
                    + all_column_infos.on_demand_column_infos
    ]
    # Excluded columns that are on-demand inputs or feature lookup keys
    # should still be in feature_spec.yaml with include=False.
    on_demand_input_names = common_utils.get_unique_list_order(
        [
            input_name
            for odci in all_column_infos.on_demand_column_infos
            for input_name in odci.input_bindings.values()
        ]
    )
    lookup_keys_and_on_demand_inputs = set(on_demand_input_names)
    for fci in all_column_infos.feature_column_infos:
        lookup_keys_and_on_demand_inputs.update(fci.lookup_key)

    column_infos = [
        ci
        for ci in column_infos
        if ci.include or ci.output_name in lookup_keys_and_on_demand_inputs
    ]

    # Sort table_infos by table_name, function_infos by udf_name, so they appear sorted in feature_spec.yaml
    # Exclude unnecessary table_infos, function_infos from the FeatureSpec. When a FeatureLookup or FeatureFunction
    # output feature is excluded, the underlying table or UDF is not required in the FeatureSpec.
    consumed_table_names = [
        ci.info.table_name
        for ci in column_infos
        if isinstance(ci.info, FeatureColumnInfo)
    ]
    consumed_table_names = sorted(set(consumed_table_names))
    consumed_udf_names = [
        ci.info.udf_name
        for ci in column_infos
        if isinstance(ci.info, OnDemandColumnInfo)
    ]
    consumed_udf_names = sorted(set(consumed_udf_names))

    # Collect lookback windows
    table_lookback_window_map = _validate_and_convert_lookback_windows(feature_lookups)

    table_infos = [
        FeatureTableInfo(
            table_name=table_name,
            table_id=ft_metadata.feature_table_metadata_map[table_name].table_id,
            lookback_window=table_lookback_window_map[table_name],
        )
        for table_name in consumed_table_names
    ]
    function_infos = [
        FunctionInfo(full_name=udf_name) for udf_name in consumed_udf_names
    ]

    # Build FeatureSpec
    feature_spec = FeatureSpec(
        column_infos=assign_topological_ordering(
            column_infos=column_infos,
        ),
        table_infos=table_infos,
        function_infos=function_infos,
        serialization_version=FeatureSpec.SERIALIZATION_VERSION_NUMBER,
    )

    return feature_spec


def add_inferred_source_columns(column_infos):
    on_demand_input_names = common_utils.get_unique_list_order(
        [
            input_name
            for odci in column_infos.on_demand_column_infos
            for input_name in odci.input_bindings.values()
        ]
    )
    on_demand_output_names = [
        odci.output_name for odci in column_infos.on_demand_column_infos
    ]

    feature_input_names = common_utils.get_unique_list_order(
        [
            input_name
            for fci in column_infos.feature_column_infos
            for input_name in fci.lookup_key
        ]
    )
    feature_output_names = [
        fci.output_name for fci in column_infos.feature_column_infos
    ]

    all_output_names = feature_output_names + on_demand_output_names

    missing_lookup_inputs = [
        feature_input_name
        for feature_input_name in feature_input_names
        if feature_input_name not in all_output_names
    ]
    missing_on_demand_inputs = [
        on_demand_input_name
        for on_demand_input_name in on_demand_input_names
        if on_demand_input_name not in all_output_names
    ]

    inferred_inputs = common_utils.get_unique_list_order(
        missing_lookup_inputs + missing_on_demand_inputs
    )
    source_data_column_infos = [SourceDataColumnInfo(col) for col in inferred_inputs]

    return _ColumnInfos(
        source_data_column_infos=column_infos.source_data_column_infos
                                 + source_data_column_infos,
        feature_column_infos=column_infos.feature_column_infos,
        on_demand_column_infos=column_infos.on_demand_column_infos,
    )



# -*- coding: utf-8 -*-

__doc__ = """
Feast客户端，用于与Feast服务器交互
"""

import logging
import json
import os
from typing import Union, List, Dict, Optional
from feast import FeatureStore, RepoConfig, FeatureView
from pyspark.sql import DataFrame, SparkSession
from wedata.feature_store.common.store_config.redis import RedisStoreConfig
from feast import Entity, Field, FileSource
from feast.data_source import RequestSource
from pyspark.sql.functions import current_timestamp
from feast.types import ValueType
from pyspark.sql.types import (
    StringType, IntegerType, LongType, FloatType, DoubleType,
    BooleanType, BinaryType, TimestampType, DateType, DecimalType,
    ArrayType, MapType, StructType, NullType, StructField
)

TEMP_FILE_PATH = "/tmp/feast_data/"


class FeastClient:

    def __init__(self, offline_store: SparkSession, online_store_config: RedisStoreConfig = None):
        project_id = os.getenv("WEDATA_PROJECT_ID", "")
        remote_path = os.getenv("FEAST_REMOTE_ADDRESS", "")
        if offline_store is None or not isinstance(offline_store, SparkSession):
            raise ValueError("offline_store must be provided SparkSession instance")

        # 应用Spark配置
        spark_conf_dict = dict()
        spark_conf = offline_store.sparkContext.getConf().getAll()
        for item in spark_conf:
            print(f"type:{type(item)} value:{item}")
            spark_conf_dict[item[0]] = item[1]

        print("spark_conf:\n", spark_conf_dict)
        config = RepoConfig(
            project=project_id,
            registry={"registry_type": "remote", "path": remote_path},
            provider="local",
            online_store={"type": "redis",
                          "connection_string": online_store_config.connection_string} if online_store_config else None,
            offline_store={"type": "spark", "spark_conf": spark_conf_dict},
            batch_engine={"type": "spark.engine"}
        )

        self._client = FeatureStore(config=config)
        self._spark = offline_store


    @property
    def client(self):
        return self._client

    def create_table(self,
                     table_name: str,
                     primary_keys: List[str],
                     timestamp_keys: Union[str, List[str]],
                     df: Optional[DataFrame] = None,
                     schema: Optional[StructType] = None,
                     description: Optional[str] = None,
                     tags: Optional[Dict[str, str]] = None):
        if schema is not None and df is None:
            # 创建空的Spark DataFrame
            df = self._spark.createDataFrame([], schema)
        feature_view = _create_table_to_feature_view(
            table_name=table_name,
            primary_keys=primary_keys,
            timestamp_keys=timestamp_keys,
            df=df,
            tags=tags
        )
        self._client.apply(feature_view)

    def remove_offline_table(self, table_name: str):
        self._client.delete_feature_view(table_name)

    def write_table(self, table_name: str, df: DataFrame):
        try:
            logging.info(f"Starting to write table {table_name}")

            # 确保临时目录存在
            os.makedirs(TEMP_FILE_PATH, exist_ok=True)

            # 获取Spark DataFrame的schema信息
            schema = df.schema

            # 通过schema识别时间戳列
            timestamp_cols = [field.name for field in schema.fields
                              if isinstance(field.dataType, (TimestampType, DateType))]

            if not timestamp_cols:
                raise ValueError("No timestamp columns found in DataFrame schema")

            event_timestamp = timestamp_cols[0]  # 默认使用第一个时间戳列作为event_timestamp
            created_timestamp = timestamp_cols[-1] if len(timestamp_cols) > 1 else None

            logging.info(f"Detected timestamp columns from schema: {timestamp_cols}")
            logging.info(f"Using event_timestamp: {event_timestamp}")
            if created_timestamp:
                logging.info(f"Using created_timestamp: {created_timestamp}")

            # 转换为Pandas DataFrame
            logging.info("Converting Spark DataFrame to Pandas")
            pd_df = df.toPandas()

            # 写入Parquet文件
            file_path = os.path.join(TEMP_FILE_PATH, f"{table_name}.parquet")
            logging.info(f"Writing to parquet file at {file_path}")
            pd_df.to_parquet(file_path, engine='pyarrow')

            # 创建FileSource
            file_source = FileSource(
                name=table_name,
                path=file_path,
                event_timestamp_column=event_timestamp,
                created_timestamp_column=created_timestamp,
                date_partition_column="date"  # 可选
            )

            # 获取现有的FeatureView
            logging.info(f"Getting feature view {table_name}")
            feature_view = self._client.get_feature_view(table_name)

            # 更新FeatureView的source
            logging.info("Updating feature view source")
            feature_view.source = file_source

            # 应用更新
            logging.info("Applying changes to Feast")
            self._client.apply([feature_view])

            logging.info("Successfully updated feature view")
        except Exception as e:
            logging.error(f"Failed to write table {table_name}: {str(e)}")
            raise

    def alter_table(self, full_table_name: str):
        """
        将已注册的Delta表同步到Feast中作为离线特征数据
        
        Args:
            full_table_name: 表名（格式：<table>）
        Raises:
            ValueError: 当表不存在或参数无效时抛出
            RuntimeError: 当同步操作失败时抛出
        """
        import logging
        try:

            # 构建完整表名

            logging.info(f"Starting to sync Delta table {full_table_name} to Feast")

            # 1. 读取Delta表数据和schema
            df = self._spark.table(full_table_name)

            # 2. 从表属性中获取主键和时间戳列
            tbl_props = self._spark.sql(f"SHOW TBLPROPERTIES {full_table_name}").collect()
            props = {row['key']: row['value'] for row in tbl_props}

            primary_keys = props.get("primaryKeys", "").split(",")
            timestamp_keys = props.get("timestampKeys", "").split(",")

            if not primary_keys:
                raise ValueError("Primary keys not found in table properties")
            if not timestamp_keys:
                raise ValueError("Timestamp keys not found in table properties")

            logging.info(f"Primary keys: {primary_keys}")
            logging.info(f"Timestamp keys: {timestamp_keys}")

            # 3. 创建或更新FeatureView
            feature_view = _create_table_to_feature_view(
                table_name=full_table_name,
                primary_keys=primary_keys,
                timestamp_keys=timestamp_keys,
                df=df,
                tags={"source": "delta_table", **json.loads(props.get("tags", "{}"))}
            )

            # 4. 应用到Feast
            self._client.apply(feature_view)
            logging.info(f"Successfully synced Delta table {full_table_name} to Feast")

        except Exception as e:
            logging.error(f"Failed to sync Delta table to Feast: {str(e)}")
            raise RuntimeError(f"Failed to sync Delta table {full_table_name} to Feast: {str(e)}") from e

    def modify_tags(
            self,
            table_name: str,
            tags: Dict[str, str]
    ) -> None:
        """修改特征表的标签信息

        Args:
            table_name: 特征表名称(格式: <database>.<table>)
            tags: 要更新的标签字典

        Raises:
            ValueError: 当参数无效时抛出
            RuntimeError: 当修改操作失败时抛出
        """
        if not table_name:
            raise ValueError("table_name cannot be empty")
        if not tags:
            raise ValueError("tags cannot be empty")

        try:
            # 获取现有的FeatureView
            feature_view = self._client.get_feature_view(table_name)
            if not feature_view:
                raise ValueError(f"FeatureView '{table_name}' not found")

            # 更新标签
            current_tags = feature_view.tags or {}
            current_tags.update(tags)
            feature_view.tags = current_tags

            # 应用更新
            self._client.apply([feature_view])
            print(f"Successfully updated tags for table '{table_name}'")

        except Exception as e:
            raise RuntimeError(f"Failed to modify tags for table '{table_name}': {str(e)}") from e


def _create_table_to_feature_view(
        table_name: str,
        primary_keys: List[str],
        timestamp_keys: Union[str, List[str]],
        df: Optional[DataFrame],
        tags: Optional[Dict[str, str]] = None
):
    """

    Returns:
        FeatureView实例
    """

    entities = list()
    for primary_key in primary_keys:
        entities.append(Entity(name=primary_key, value_type=ValueType.STRING))

    if primary_keys is None or len(primary_keys) == 0:
        raise ValueError("primary_keys must not be empty")
    if timestamp_keys is None or len(timestamp_keys) == 0:
        raise ValueError("timestamp_keys must not be empty")
        
    # 处理timestamp_keys
    missing_timestamps = []
    if isinstance(timestamp_keys, str):
        if timestamp_keys not in df.columns:
            df = df.withColumn(timestamp_keys, current_timestamp())
            missing_timestamps.append(timestamp_keys)
    elif isinstance(timestamp_keys, list):
        for ts_key in timestamp_keys:
            if ts_key not in df.columns:
                df = df.withColumn(ts_key, current_timestamp())
                missing_timestamps.append(ts_key)

    if missing_timestamps:
        print(f"Added missing timestamp columns: {missing_timestamps}")

    os.makedirs(TEMP_FILE_PATH, exist_ok=True)
    df.toPandas().to_parquet(os.path.join(TEMP_FILE_PATH, f"{table_name}.parquet"), engine='pyarrow')

    resources = FileSource(
        name=table_name,
        path=os.path.join(TEMP_FILE_PATH, f"{table_name}.parquet"),
        event_timestamp_column=timestamp_keys[0] if isinstance(timestamp_keys, list) else timestamp_keys,
        created_timestamp_column=timestamp_keys[-1] if isinstance(timestamp_keys, list) and len(timestamp_keys) > 1 else None,
        owner="",
    )

    # 构建FeatureView的剩余逻辑
    feature_view = FeatureView(
        name=table_name,
        entities=entities,
        tags=tags,
        source=resources,
    )

    return feature_view



import copy
import datetime
import logging
from typing import Dict, List, Optional, Union

from wedata.feature_store.utils import common_utils

_logger = logging.getLogger(__name__)


class FeatureLookup:

    """
    特征查找类

    特征查找类用于指定特征表中的特征，并将其与训练集中的特征进行关联。

    特征查找类有以下属性：

    - table_name：特征表的名称。
    - lookup_key：用于在特征表和训练集之间进行联接的键。lookup_key必须是训练集中的列。lookup_key的类型和顺序必须与特征表的主键匹配。
    - feature_names：要从特征表中查找的特征的名称。如果您的模型需要主键作为特征，则可以将它们声明为独立的FeatureLookups。
    - rename_outputs：如果提供，则会将特征重命名为 :meth:`create_training_set() <databricks.feature_engineering.client.FeatureEngineeringClient.create_training_set>`返回的 :class:`TrainingSet <databricks.ml_features.training_set.TrainingSet>` 中的特征。
    - timestamp_lookup_key：用于在特征表和训练集之间进行联接的时间戳键。timestamp_lookup_key必须是训练集中的列。timestamp_lookup_key的类型必须与特征表的时间戳键的类型匹配。
    - lookback_window: 当对特征表执行时间点查找时使用的回溯窗口，该查找针对传递给 :meth:`create_training_set() <databricks.feature_engineering.client.FeatureEngineeringClient.create_training_set>` 方法的数据帧。特征存储将检索在数据帧的``timestamp_lookup_key``指定时间戳之前且在``lookback_window``时间范围内的最新特征值，如果不存在这样的特征值则返回null。当设置为0时，仅返回特征表中的精确匹配项。
    - feature_name：特征名称。**已弃用**。使用 `feature_names`。
    - output_name：如果提供，则会将此特征重命名为 :meth:`create_training_set() <databricks.feature_engineering.client.FeatureEngineeringClient.create_training_set>` 返回的 :class:`TrainingSet <databricks.ml_features.training_set.TrainingSet>` 中的特征。**已弃用**。使用 `rename_outputs`。

    示例：

    from databricks.feature_store import FeatureLookup

    lookup = FeatureLookup(
        table_name="my_feature_table",
        lookup_key="my_lookup_key",
        feature_names=["my_feature_1", "my_feature_2"],
        rename_outputs={"my_feature_1": "my_feature_1_renamed"},
        timestamp_lookup_key="my_timestamp_lookup_key",
        lookback_window=datetime.timedelta(days=1)
    )

    """

    def __init__(
        self,
        table_name: str,
        lookup_key: Union[str, List[str]],
        *,
        feature_names: Union[str, List[str], None] = None,
        rename_outputs: Optional[Dict[str, str]] = None,
        timestamp_lookup_key: Optional[str] = None,
        lookback_window: Optional[datetime.timedelta] = None,
        **kwargs,
    ):
        """Initialize a FeatureLookup object. See class documentation."""

        self._feature_name_deprecated = kwargs.pop("feature_name", None)
        self._output_name_deprecated = kwargs.pop("output_name", None)

        if kwargs:
            raise TypeError(
                f"FeatureLookup got unexpected keyword argument(s): {list(kwargs.keys())}"
            )

        self._table_name = table_name

        if type(timestamp_lookup_key) is list:
            if len(timestamp_lookup_key) == 0:
                timestamp_lookup_key = None
            elif len(timestamp_lookup_key) == 1:
                timestamp_lookup_key = timestamp_lookup_key[0]
            else:
                raise ValueError(
                    f"Setting multiple timestamp lookup keys is not supported."
                )

        if rename_outputs is not None and not isinstance(rename_outputs, dict):
            raise ValueError(
                f"Unexpected type for rename_outputs: {type(rename_outputs)}"
            )

        self._feature_names = common_utils.as_list(feature_names, default=[])

        # Make sure the user didn't accidentally pass in any nested lists/dicts in feature_names
        for fn in self._feature_names:
            if not isinstance(fn, str):
                raise ValueError(
                    f"Unexpected type for element in feature_names: {type(self._feature_names)}, only strings allowed in list"
                )

        if lookback_window is not None:
            if not timestamp_lookup_key:
                raise ValueError(
                    f"Unexpected lookback_window value: {lookback_window}, lookback windows can only be applied on time series "
                    f"feature tables. Use timestamp_lookup_key to perform point-in-time lookups with lookback window."
                )
            if not isinstance(
                lookback_window, datetime.timedelta
            ) or lookback_window < datetime.timedelta(0):
                raise ValueError(
                    f"Unexpected value for lookback_window: {lookback_window}, only non-negative datetime.timedelta allowed."
                )

        self._lookup_key = copy.copy(lookup_key)
        self._timestamp_lookup_key = copy.copy(timestamp_lookup_key)
        self._lookback_window = copy.copy(lookback_window)

        self._rename_outputs = {}
        if rename_outputs is not None:
            self._rename_outputs = rename_outputs.copy()

        self._inject_deprecated_feature_name()
        self._inject_deprecated_output_name()

    @property
    def table_name(self):
        """The table name to use in this FeatureLookup."""
        return self._table_name

    @property
    def lookup_key(self):
        """The lookup key(s) to use in this FeatureLookup."""
        return self._lookup_key

    @property
    def feature_name(self):
        """The feature name to use in this FeatureLookup. **Deprecated**. Use `feature_names`."""
        return self._feature_name_deprecated

    @property
    def feature_names(self):
        """The feature names to use in this FeatureLookup."""
        return self._feature_names

    @property
    def output_name(self):
        """The output name to use in this FeatureLookup. **Deprecated**. Use `feature_names`."""
        if self._output_name_deprecated:
            return self._output_name_deprecated
        else:
            return self._feature_name_deprecated

    @property
    def timestamp_lookup_key(self):
        return self._timestamp_lookup_key

    @property
    def lookback_window(self):
        """A lookback window applied only for point-in-time lookups."""
        return self._lookback_window

    def _get_feature_names(self):
        return self._feature_names

    def _get_output_name(self, feature_name):
        """Lookup the renamed output, or fallback to the feature name itself if no mapping is present"""
        return self._rename_outputs.get(feature_name, feature_name)

    def _inject_deprecated_feature_name(self):
        if self._feature_name_deprecated:
            if len(self._feature_names) > 0:
                raise ValueError(
                    "Use either feature_names or feature_name parameter, but not both."
                )
            _logger.warning(
                f'The feature_name parameter is deprecated. Use "feature_names".'
            )
            self._feature_names = [self._feature_name_deprecated]

    def _inject_deprecated_output_name(self):
        if len(self._feature_names) == 1 and self._output_name_deprecated:
            if len(self._rename_outputs) > 0:
                raise ValueError(
                    "Use either output_name or rename_outputs parameter, but not both."
                )
            _logger.warning(
                f'The output_name parameter is deprecated.  Use "rename_outputs".'
            )
            self._rename_outputs[self._feature_names[0]] = self._output_name_deprecated

    @table_name.setter
    def table_name(self, value):
        self._table_name = value
